// ########################################################################
// ## liste.cc:                                                          ##
// ## Liste zusammenhngender, trivalenter Graphen                       ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ##           Arash Shahi,      shahi@xmailer.informatik.uni-bonn.de   ##
// ##           Jamshid Azizi,    azizi@xmailer.informatik.uni-bonn.de   ##
// ########################################################################

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <bool.h>
#include <minmax.h>

#include "trigraph.h"
#include "trigraph.icc"
#include "trigraphkatalog.h"
#include "trigraphkatalog.icc"


TriGraphKatalog* katalog;
TriGraphListe*   start;


// Ausgabe-Einstellungen
int TriGraph::ausgabeEinstellungen = TriGraph::ausgabeANSI;


// ########################################################################
// ## Harary-Startbume                                                  ##
// ########################################################################

const unsigned int cBaumEnde  = 4000000000;
const unsigned int cListeEnde = 4000000001;

unsigned int StartBaum1[] = {
   cListeEnde,cListeEnde
};
unsigned int StartBaum2[] = {
   1,2, cListeEnde,cListeEnde
};
unsigned int StartBaum3[] = {
   1,2, 2,3, cListeEnde,cListeEnde,
};
unsigned int StartBaum4[] = {
   1,2, 2,3, 3,4, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, cListeEnde,cListeEnde
};
unsigned int StartBaum5[] = {
   1,2, 2,3, 3,4, 4,5, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, cListeEnde,cListeEnde
};
unsigned int StartBaum6[] = {
   1,2, 2,3, 3,4, 4,5, 5,6, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 2,5, 5,6, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 4,6, cListeEnde,cListeEnde
};
unsigned int StartBaum7[] = {
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 5,7, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 4,7, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 3,6, 6,7, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 5,7, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 4,7, cListeEnde,cListeEnde
};
unsigned int StartBaum8[] = {
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 6,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 5,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 4,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 2,6, 6,7, 7,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 6,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 5,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 4,7, 7,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 5,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 4,8, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 4,6, 6,7, 6,8, cListeEnde,cListeEnde
};
unsigned int StartBaum9[] = {
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 7,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 6,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 5,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 5,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 3,7, 7,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 7,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 6,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 5,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 6,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 5,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 4,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 4,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 5,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 5,8, 8,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 2,5, 5,6, 5,7, 7,8, 7,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 5,7, 7,8, 7,9, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 4,7, 7,8, 7,9, cListeEnde, cListeEnde
};
unsigned int StartBaum10[] = {
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 8,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 8,9, 7,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 8,9, 6,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 6,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 2,7, 7,8, 8,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 7,8, 5,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 6,7, 4,8, 8,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 6,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 8,9, 7,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 8,9, 6,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 8,9, 5,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 5,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 8,9, 6,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 6,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 8,9, 4,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 7,8, 4,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 7,8, 8,9, 5,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 6,7, 4,8, 8,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, 6,7, 7,8, 8,9, 6,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 5,8, 8,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 3,6, 6,7, 7,8, 8,9, 6,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 3,6, 6,7, 7,8, 6,9, 9,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 5,7, 4,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, 6,7, 3,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 6,7, 2,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 3,6, 6,7, 2,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 3,5, 5,6, 2,7, 7,8, 8,9, 7,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 5,7, 3,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, 6,7, 2,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, 2,7, 7,8, 8,9, 7,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 5,6, 5,7, 2,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 3,4, 4,5, 4,6, 2,7, 7,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 5,6, 5,7, 4,8, 8,9, 8,10, cBaumEnde, cBaumEnde,
   1,2, 2,3, 2,4, 4,5, 4,6, 6,7, 6,8, 8,9, 8,10, cListeEnde, cListeEnde 
};

unsigned int* StartBaeume[] = {
   { NULL},
   { StartBaum1 },
   { StartBaum2 },
   { StartBaum3 },
   { StartBaum4 },
   { StartBaum5 },
   { StartBaum6 },
   { StartBaum7 },
   { StartBaum8 },
   { StartBaum9 },
   { StartBaum10 }
};


// ########################################################################
// ## Graphen-Erzeugung                                                  ##
// ########################################################################


// Alle Mglichkeiten durchprobieren
void Erzeuge(TriGraph& graph, const unsigned int N,
       const unsigned int* quelle, const unsigned int* ziel,
       unsigned int anzahl, unsigned int zaehler)
{
   // ===== Kanten systematisch hinzufgen ================================
   if(zaehler < anzahl) {
      // Erzeugung mit Mglichkeit quelle[zhler] und ziel[zhler] getrennt
      Erzeuge(graph,N,quelle,ziel,anzahl,zaehler + 1);

      // Erzeugung mit Mglichkeit quelle[zhler] und ziel[zhler] verbunden
      if( graph.Verbindbar(quelle[zaehler],ziel[zaehler]) ) {
         graph.Verbinde(quelle[zaehler],ziel[zaehler]);
         Erzeuge(graph,N,quelle,ziel,anzahl,zaehler + 1);
         graph.Trenne(quelle[zaehler],ziel[zaehler]);
      } 
   }
   else {
      katalog->FuegeEin(graph);
   }
}


// Graphen zu Baum erzeugen 
void ErzeugeGraphenZuBaum(const TriGraph& baum, const unsigned int N)
{
   TriGraph graph = baum;
   katalog->FuegeEin(baum);              

   // ===== Menge der mglichen neuen Kanten ermitteln ====================
   unsigned int quelle[N * N];
   unsigned int ziel[N * N];
   unsigned int anzahl = 0;
   for(unsigned int x = 0;x < N;x++) {
      if(graph.Grad(x) < 3) {
         for(unsigned int y = x + 1;y < N;y++) {
            if(graph.Getrennt(x,y) && (graph.Grad(y) < 3)) {          
               quelle[anzahl] = x; 
               ziel[anzahl]   = y;
               anzahl++;
             }
         }
      }
   }

   // ===== Zum Baum passende Graphen erzeugen ============================
   Erzeuge(graph,N,quelle,ziel,anzahl,0);
}


// Graphen erzeugen
void ErzeugeGraphen(const unsigned int N)
{
   TriGraph baum;
   int      k  = 0;
   bool     ok = start->Erster(baum);

   while(ok == true) {
      cerr << "Erzeuge Graphen zu Startbaum #" << ++k << "..." << endl;
      ErzeugeGraphenZuBaum(baum,N);
      ok = start->Nachfolger(baum);      
   }       
}


// Startbume mit N Knoten erzeugen
void ErzeugeStartbaeume(const unsigned int N)
{
   if((N == 0) || (N > 10)) {
      cerr << "Fehler: Fr angegebenes N liegen keine Start-Bume vor!" << endl;
      exit(1);
   }

   TriGraph graph(N),null(N);
   unsigned int* baum = StartBaeume[N];
   unsigned int k = 0;

   cerr << "Erzeuge Startbume..." << endl;
   while(baum[k] != cListeEnde) {
      if(baum[k] != cBaumEnde) {
         graph.Verbinde(baum[k] - 1,baum[k + 1] - 1);
      }
      else {
         start->FuegeAnOhneVergleich(graph);
         graph = null;
      }
      k += 2;
   }
   start->FuegeAnOhneVergleich(graph); 
}


// ########################################################################
// ## Hauptprogramm                                                      ##
// ########################################################################


int main(int argc, char** argv)
{
   if(argc < 2) {
      cerr << "Aufruf: " << argv[0] << " [Knoten] {Dateiname}" << endl;
      exit(1);
   }
   unsigned int N = atol(argv[1]);
   cout << "Knoten = " << N << endl << endl;

   // Katalog und Startbaum-Liste erzeugen   
   katalog = new TriGraphKatalog(N);
   start   = new TriGraphListe;
   if((katalog == NULL) || (start == NULL)) {
      cerr << "Nicht genug Speicher fr Katalog!" << endl;
      exit(1);
   }  

   // Graphen erzeugen
   ErzeugeStartbaeume(N);
   ErzeugeGraphen(N);

   // Graphen ausgeben
   if(argc > 2) {
      ofstream os(argv[2],ios::binary | ios::out);
      if(os.good()) {
         katalog->Schreibe(os);
         cout << "Katalog erfolgreich gespeichert!" << endl;
      } 
      else {
         cerr << "Fehler beim ffnen der Ausgabe-Datei!" << endl;
      }
   }
   else {
      cout << *katalog << endl;
   }

   // Programmende
   delete start;
   delete katalog;
   return 0;
}
