// ########################################################################
// ## trigraph.cc:                                                       ##
// ## Klasse fr zusammenhngende, trivalente Graphen                    ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ##           Arash Shahi,      shahi@xmailer.informatik.uni-bonn.de   ##
// ##           Jamshid Azizi,    azizi@xmailer.informatik.uni-bonn.de   ##
// ########################################################################

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <bool.h>
#include <minmax.h>

#include "trigraph.h"
#include "trigraph.icc"
#include "isotest.h"


// ########################################################################
// ## Konstruktoren und Destruktor                                       ##
// ########################################################################


// Konstruktor 0
TriGraph::TriGraph()
{
   for(int i = 0;i < 3;i++) {
      gradPartition[i] = 0;
   }
   adjazenzMatrix = NULL;
   nachfolger     = NULL;
   knoten         = 0;
}


// Klasse initialisieren
void TriGraph::Initialisiere(const unsigned int n)
{
   unsigned int i;

   knoten     = (unsigned short)n;
   kanten     = 0;
   nachfolger = NULL;
   for(i = 0;i < 3;i++) {
      gradPartition[i] = 0;
   }
   adjazenzMatrix = new unsigned char[AdjazenzBytes(n)];
   if(adjazenzMatrix == NULL) {
      cerr << "Nicht genug freier Speicher!" << endl;
      exit(1);
   }
   for(i = 0;i < AdjazenzBytes(n);i++) {
      adjazenzMatrix[i] = 0;
   }
}


// Klasse freigeben
void TriGraph::GibFrei()
{
   if(adjazenzMatrix != NULL) {
      delete adjazenzMatrix;
      adjazenzMatrix = NULL;
      nachfolger     = NULL;
   } 
}


// Konstruktor 1
TriGraph::TriGraph(const unsigned int n)
{  
   Initialisiere(n);
}


// Konstruktor 2
TriGraph::TriGraph(const TriGraph& original)
{  
   Clone(original);
}


// Copy-Konstruktor
void TriGraph::operator=(const TriGraph& original)
{  
   GibFrei();
   Clone(original);
}


// Clone-Operation
void TriGraph::Clone(const TriGraph& original)
{
   unsigned int i;

   Initialisiere(original.knoten);
   kanten = original.kanten;
   for(i = 0;i < 3;i++) {
      gradPartition[i] = original.gradPartition[i];
   }   
   for(i = 0;i < AdjazenzBytes(original.knoten);i++) {
      adjazenzMatrix[i] = original.adjazenzMatrix[i];
   }
}


// Destruktor
TriGraph::~TriGraph()
{
   GibFrei();
}


// ########################################################################
// ## Verwaltung der Adjazenzmatrix                                      ##
// ########################################################################


// Kante zwischen Knoten a und b hinzufgen oder lschen
void TriGraph::SetzeKante(const unsigned int a, const unsigned int b,
                          const bool neueKante)
{
#ifdef DEBUG
   if((a >= knoten) || (b >= knoten)) {
      cerr << "TriGraph::SetzeKante() - Unzulssige Knotennummern!" << endl;
      exit(1);
   }
   if(a == b) {
      cerr << "TriGraph::SetzeKante() - Schleifen sind nicht erlaubt!" << endl;
      exit(1);
   }
   if(Verbunden(a,b) == neueKante) {
      cerr << "TriGraph::Kante() - Befehl unntig!" << endl;
      exit(1);
    }
#endif

#ifdef BIT_MATRIX
   // Bitposition in der Adjazenzmatrix berechnen
   const int z         = min(a,b);
   const int s         = max(a,b);
   const int bitNummer = (z * knoten) + s - ((z + 2) * (z + 1)) / 2;
   const int byte      = bitNummer / 8;
   const int bit       = bitNummer % 8;

   // Neue Kante hinzufgen bzw. entfernen
   if(neueKante) {
      if(! (adjazenzMatrix[byte] & (1 << bit))) {
         adjazenzMatrix[byte] += (1 << bit);
         kanten++;
      }
   }
   else {
      if(adjazenzMatrix[byte] & (1 << bit)) {
         adjazenzMatrix[byte] -= (1 << bit);
         kanten--;
      }
   }
#else
   const int z         = min(a,b);
   const int s         = max(a,b);
   adjazenzMatrix[(z * knoten) + s - (((z + 2) * (z + 1)) / 2)] = (char)neueKante;
   if(neueKante)
      kanten++;
   else
      kanten--;
#endif

   // Grad-Partition aktualisieren
   SetzeGradpartition();
}


// Grad-Partition fr Graphen berechnen
void TriGraph::SetzeGradpartition()
{
   // Alte Werte lschen
   for(int i = 0;i < 3;i++) {
      gradPartition[i] = 0;
   }

   // Von jedem Knoten den Grad berechnen und die der Gradzahl entsprechende
   // Komponente der Gradpartition erhhen.
   for(int i = 0;i < knoten;i++) {
      int grad = Grad(i);
#ifdef DEBUG
      if(grad > 3) {
         cerr << "TriGraph::SetzeGradpartition() - Grad > 3!" << endl;
         exit(1);
      }
#endif
      if(grad != 0) {
         gradPartition[grad - 1]++;
      }
   }
}


// ########################################################################
// ## Ausgabe                                                            ##
// ########################################################################


// Output-Operator
ostream& operator<<(ostream& os, const TriGraph& graph)
{
   if(graph.knoten > 0) {
      if(!(graph.ausgabeEinstellungen & TriGraph::ausgabeKompakt)) {
         os << "Knoten   = " << graph.knoten << endl; 
         os << "Kanten   = " << graph.kanten << endl;
         os << "GradPart = " << graph.gradPartition[0] << " "
                             << graph.gradPartition[1] << " "
                             << graph.gradPartition[2] << endl;
         os << "AdMatrix =   ";
         for(int i = 0;i < graph.knoten;i++) {
            os.width(3);
            os << i;
         }
         os << endl << "          ";
         for(int zeile = 0;zeile < graph.knoten;zeile++) {
            os.width(3);
            os << zeile << " ";
            for(int spalte = 0;spalte < graph.knoten;spalte++) {
               if(graph.ausgabeEinstellungen & TriGraph::ausgabeANSI) {
                  if(zeile == spalte) 
                     os << "\x1b[0;36m - \x1b[0;39m";
                  else if(graph.Verbunden(zeile,spalte))
                     os << "\x1b[1;33m 1 \x1b[0;39m";
                  else
                     os << "\x1b[1;34m 0 \x1b[0;39m";
               }
               else {
                  if(zeile == spalte) 
                     os << " - ";
                  else if(graph.Verbunden(zeile,spalte))
                     os << " 1 ";
                  else
                     os << " 0 ";
               }
            }
            os << endl;
            if(zeile < graph.knoten - 1) {
               os << "          ";
            }       
         }
      }
      os << "Verbindg = ";    
      for(unsigned int i = 0; i < graph.knoten;i++) {
         for(unsigned int j = i + 1; j < graph.knoten;j++) {
            if(graph.Verbunden(i,j)) {
               os << i << "-" << j << "  ";       
            }
         }
      }
      os << endl;                            
   }
   else { 
      os << "( leerer Graph )" << endl;
   }

   return(os);
}



// ########################################################################
// ## Lesen und Schreiben                                                ##
// ########################################################################


#ifdef BIT_MATRIX
const unsigned short magicID = 0x7465;
#else
const unsigned short magicID = 0x7466;
#endif


// Graph aus Binrdatei lesen
bool TriGraph::Lese(istream& is)
{
   // ID-Word lesen und vergleichen
   unsigned short id = 0;
   is.read((void *)&id,sizeof(magicID));
   if(id != magicID) {
      return(false);
   }

   // Graph lesen
   is.read((void *)&knoten,sizeof(kanten));
   is.read((void *)&kanten,sizeof(kanten));
   is.read((void *)&gradPartition,sizeof(gradPartition));
   if(adjazenzMatrix)
      delete adjazenzMatrix;
   adjazenzMatrix = new unsigned char[AdjazenzBytes(knoten)];
   if(adjazenzMatrix == NULL) {
      cerr << "TriGraph::Lese() - Nicht genug freier Speicher!" << endl;
      exit(1);
   }
   is.read((void *)adjazenzMatrix,AdjazenzBytes(knoten));  
   if(is.fail()) {
      return(false);
   }
   return(true);   
}


// Graph in Binrdatei schreiben
bool TriGraph::Schreibe(ostream& os)
{
   os.write((void *)&magicID,sizeof(magicID));   
   os.write((void *)&knoten,sizeof(knoten));   
   os.write((void *)&kanten,sizeof(kanten));   
   os.write((void *)&gradPartition,sizeof(gradPartition));   
   os.write((void *)adjazenzMatrix,AdjazenzBytes(knoten));
   if(os.fail()) {
      return(false);
   }
   return(true);   
}
