// ########################################################################
// ## trigraphliste.cc:                                                  ##
// ## Liste zusammenhngender, trivalenter Graphen                       ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ##           Arash Shahi,      shahi@xmailer.informatik.uni-bonn.de   ##
// ##           Jamshid Azizi,    azizi@xmailer.informatik.uni-bonn.de   ##
// ########################################################################


#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <bool.h>


#include "einstellungen.h"
#include "trigraph.h"
#include "trigraph.icc"
#include "trigraphliste.h"
#include "trigraphliste.icc"


// ########################################################################
// ## Konstruktoren und Destruktor                                       ##
// ########################################################################
 
 
// Konstruktor 1
TriGraphListe::TriGraphListe()
{
   Initialisiere();
} 


// Klasse initialisieren
void TriGraphListe::Initialisiere()
{
   elemente  = 0;
   erster    = NULL;
   letzter   = NULL;
   aktueller = NULL;
}


// Klasse freigeben
void TriGraphListe::GibFrei()
{
   if(erster != NULL) {
      elemente  = 0;
      erster    = NULL;   
      letzter   = NULL;
      aktueller = NULL;
   }
}


// Clone-Operation
void TriGraphListe::Clone(const TriGraphListe& original)
{
   Initialisiere();
   TriGraph* t = original.erster;
   elemente    = original.elemente;
   while(t != NULL) {
      TriGraph* neu = new TriGraph(*t);
      if(letzter != NULL) {
         letzter->SetzeNachfolger(neu);
      }
      else {
         erster = neu;
      }   
      letzter = neu;      
      t = t->Nachfolger();
   }
   aktueller = erster;
}


// Konstruktor 2
TriGraphListe::TriGraphListe(const TriGraphListe& original)
{
   Clone(original);
} 


// Destruktor
TriGraphListe::~TriGraphListe()
{
   GibFrei();
}


// ########################################################################
// ## Operatoren                                                         ##
// ########################################################################


// Output-Operator
ostream& operator<<(ostream& os, const TriGraphListe& liste)
{
   if(liste.erster != NULL) {
      TriGraph* t      = liste.erster;
      unsigned int num = 1;
      while(t != NULL) {
         os << "----- ";
         os.width(4);
         os << num << " von ";
         os.width(4);
         os << liste.Elemente() << " -------------------------------------------------------" << endl;
         os << *t;
         t = t->Nachfolger();
         num++;
      }   
   }
   else {
      os << "{ leere Liste }" << endl;
   }
   return(os);
}


// =-Operator
void TriGraphListe::operator=(const TriGraphListe& original)
{
   GibFrei();
   Clone(original);
}


// ########################################################################
// ## Bearbeitung der Liste                                              ##
// ########################################################################


// Graph an Liste anfgen ohne Isomorphie-Test durchzufhren
void TriGraphListe::FuegeAnOhneVergleich(const TriGraph& t)
{
   TriGraph* neu = new TriGraph(t);
   if(letzter) {
      letzter->SetzeNachfolger(neu);
      letzter = aktueller = neu;
      elemente++;
   }
   else {
      erster = letzter = aktueller = neu;
      elemente = 1;
   }
}


// Graph an Liste anfgen, falls noch nicht in Liste vorhanden      
bool TriGraphListe::FuegeAn(const TriGraph& t)
{
   TriGraph* neu = new TriGraph(t);

   aktueller = erster;
   if(aktueller != NULL) {
      if(*neu == *aktueller) {          
         return(false);
      }
      while(aktueller->Nachfolger() != NULL) {
         aktueller = aktueller->Nachfolger();
         if(*neu == *aktueller) {          
            return(false);
         }
      }
      aktueller->SetzeNachfolger(neu);
      aktueller = letzter = neu;
      elemente++;
   }
   else {
      erster   = letzter = aktueller = neu;
      elemente = 1;
   }
   return(true);
}


// Erster Graph der Liste
bool TriGraphListe::Erster(TriGraph& t)
{
   aktueller = erster;
   if(aktueller != NULL) {
      t = *aktueller;
      return(true);
   }
   return(false);
}


// Letzter Graph der Liste
bool TriGraphListe::Letzter(TriGraph& t)
{
   aktueller = letzter;
   if(aktueller != NULL) {
      t = *aktueller;
      return(true);
   }
   return(false);
}


// Nchster Graph der Liste
bool TriGraphListe::Nachfolger(TriGraph& t)
{
   if(aktueller != NULL) {
      aktueller = aktueller->Nachfolger();
      if(aktueller != NULL) {
         t = *aktueller;
         return(true);
      }   
   }
   return(false);
}


// ########################################################################
// ## Lesen und Schreiben                                                ##
// ########################################################################


// Liste aus Binrdatei lesen
bool TriGraphListe::Lese(istream& is)
{
   GibFrei();
   for(;;) {
      TriGraph graph;
      if(!graph.Lese(is)) {
         return(true);
      }
      FuegeAnOhneVergleich(graph);
   }
   return(true);
}


// Liste in Binrdatei schreiben
bool TriGraphListe::Schreibe(ostream& os)
{
   TriGraph* graph = erster;
   while(graph != NULL) {
      if(!graph->Schreibe(os)) {
         #ifdef DEBUG
         cerr << "TriGraphListe::Schreibe() - Fehler beim Schreiben!" << endl;
         #endif
         return(false);
      }
      graph = graph->Nachfolger();
   }
   return(true);
}
